'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var INIT_REDUX = '@@redux/INIT';
var INIT_DEVTOOLS = '@@INIT';
var INIT_REDUX_OPERATIONS = '@@reduxOperations/INIT';
var REDUX_OPERATION_SIGNATURE = '@@reduxOperations';

var bindOperationToActionCreator = function bindOperationToActionCreator(locationInState, operationName, actionCreator) {
  return function () {
    var action = actionCreator.apply(undefined, arguments);
    action.meta = _extends({}, action.meta, { operations: { locationInState: locationInState, operationName: operationName } });
    return action;
  };
};

/**
 * Binds the locationInState and operationName to every actionCreator. After the action is created with the
 * enhanced function, the action will contain a meta.operations property that is used internally by redux-operations.
 * The same can be accomplished by manually passing this info into your action creators.
 *
 * @param {Array} locationInState the stack that takes you from store.getState() to your specified component state
 * @param {Function} reducer the reducer associated with this action. We just use it to grab the name
 * @param {Function|Object} actionCreators a redux action creator or object of many action creators
 *
 * @returns {Function|Object} actionCreators that automatically assign operations metadata to themselves
 */
var bindOperationToActionCreators = exports.bindOperationToActionCreators = function bindOperationToActionCreators(locationInState, reducer, actionCreators) {
  var operationName = reducer.name;
  if (typeof actionCreators === 'function') {
    return bindOperationToActionCreator(locationInState, operationName, actionCreators);
  }

  if ((typeof actionCreators === 'undefined' ? 'undefined' : _typeof(actionCreators)) !== 'object' || actionCreators === null) {
    throw new Error('bindOperationToActionCreators requires a function or object full of functions');
  }

  return Object.keys(actionCreators).reduce(function (reduction, actionCreator) {
    reduction[actionCreator] = bindOperationToActionCreator(locationInState, operationName, actionCreators[actionCreator]);
    return reduction;
  }, {});
};

/**
 * Given the redux state, walkState uses the locationInState stack as a map to get to your
 * desired subState. If it doesn't exist yet, it creates it & initializes it based on the defaultValue
 * supplied in your reducer. This is commonly used in react-redux's mapStateToProps for dynamic states.
 *
 * @param {Array} locationInState the stack that takes you from store.getState() to your specified component state
 * @param {Object} state the redux state
 * @param {Function} initializer the redux-operations reducer (not your rootReducer)
 *
 * @returns {*} the desired subState
 */
var walkState = exports.walkState = function walkState() {
  var locationInState = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
  var state = arguments[1];
  var initializer = arguments[2];

  //TODO maybe memoize?
  return locationInState.reduce(function (reduction, key, currentIdx) {
    if (reduction.hasOwnProperty(key)) {
      return reduction[key];
    } else if (currentIdx === locationInState.length - 1) {
      return typeof initializer === 'function' && initializer();
    } else {
      return {};
    }
  }, state);
};

/**
 * A custom factory to create a reducer full of operations specifically for redux-operations
 *
 * @param {String} operationName the name of the reducer as passed into combineReducers.
 * @param {*} initialState the initial state for components using this reducer, same as what you'd use for redux.
 * @param {Object} reducerObject an object where each property key is a redux actionType.
 *
 * @property {Object} actionType an object that holds the priority, resolve, and arguments of a specific operation.
 * @property {Number} actionType.priority a priority assigned to the operation. If multiple reducers share the same
 * action type, the one with a lower priority will run first. Defaults to 0.
 * @property {Function} actionType.resolve the reducer function that receives state and the
 * overloaded action (includes meta.operations) and runs on a dispatch
 * @property {Object} actionType.arguments AN object of arguments that are expected to be present in the payload
 * @property {Function} argument.type The expected constructor for the given argument
 * @property {String} argument.description A brief description of the argument passed into the payload
 *
 * @returns {Function} a redux-operations reducer that plays well with other reducers
 */
var operationReducerFactory = exports.operationReducerFactory = function operationReducerFactory(operationName, initialState, reducerObject) {
  var reducer = function reducer() {
    var state = arguments.length <= 0 || arguments[0] === undefined ? initialState : arguments[0];
    var action = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

    if (action.type !== INIT_REDUX_OPERATIONS) return state;

    //For each operation, set the initialState as the default value
    Object.keys(reducerObject).forEach(function (operation) {
      var resolveFunc = reducerObject[operation].resolve;
      reducerObject[operation].resolve = function () {
        var state = arguments.length <= 0 || arguments[0] === undefined ? initialState : arguments[0];
        var action = arguments[1];
        return resolveFunc(state, action);
      };
      reducerObject[operation].resolve.toString = function () {
        return '<Resolve Function>';
      };
      var args = reducerObject[operation].arguments;
      if ((typeof args === 'undefined' ? 'undefined' : _typeof(args)) === 'object' && args !== null) {
        Object.keys(args).forEach(function (arg) {
          var curArg = args[arg];
          if (typeof curArg.type === 'function') {
            curArg.type.toString = function () {
              return '<' + curArg.type.name + '>';
            };
          }
        });
      }
    });
    return _extends({}, reducerObject, {
      signature: REDUX_OPERATION_SIGNATURE,
      operationName: operationName
    });
  };
  return Object.defineProperty(reducer, 'name', { value: operationName });
};

/**
 * An internal function that dynamically mutates the redux state.
 *
 * @param {Array} locationInState the stack that takes you from store.getState() to your specified component state
 * @param {Object} state the redux state
 * @param {Object} newSubState the newly created subState that will overwrite the state at locationInState
 *
 * @returns {Object} new state
 */
var appendChangeToState = function appendChangeToState(locationInState, state, newSubState) {
  var nextLocation = locationInState[0];
  if (locationInState.length === 1) {
    return _extends({}, state, _defineProperty({}, nextLocation, newSubState));
  } else {
    var subObject = appendChangeToState(locationInState.slice(1), state[nextLocation] || {}, newSubState);
    return _extends({}, state, _defineProperty({}, nextLocation, subObject));
  }
};

/**
 * An internal function that generates the redux-operations API.
 *
 * @param {*} initialResult the state after initialization & populated with initial values
 *
 * @returns {Object} the beautiful visual API as seen in redux devtools
 */
var makeStoreAPI = function makeStoreAPI(initialResult) {
  var api = {};
  makeStoreOperations(api, initialResult);
  Object.keys(api).forEach(function (action) {
    api[action].operationArray.sort(function (a, b) {
      var priorityA = a.priority || 0;
      var priorityB = b.priority || 0;
      return priorityA > priorityB;
    });
    if (process.env.NODE_ENV !== 'production') {
      api[action].arguments = api[action].operationArray.reduce(function (reduction, operation) {
        return Object.assign(reduction, operation.arguments);
      }, {});
    }
  });
  return api;
};

/**
 * An internal function that picks out the reducers that are specially created
 * for redux-operations. It then provides a defaultLocation for components that don't need
 * to be dynamically created. It also provides a name that will be compared to operationName in walkState,
 * which is generally provided in the reducer or component layer.
 */
var makeStoreOperations = function makeStoreOperations(storeOperations, state) {
  var stack = arguments.length <= 2 || arguments[2] === undefined ? [] : arguments[2];
  var key = arguments[3];

  if (state && (typeof state === 'undefined' ? 'undefined' : _typeof(state)) === 'object') if (state.signature === REDUX_OPERATION_SIGNATURE) {
    Object.keys(state).filter(function (key) {
      return key !== 'signature';
    }).forEach(function (operation) {
      storeOperations[operation] = storeOperations[operation] || {};
      storeOperations[operation].operationArray = storeOperations[operation].operationArray || [];
      storeOperations[operation].operationArray.push(_extends({}, state[operation], {
        defaultLocation: [].concat(_toConsumableArray(stack)),
        name: key
      }));
    });
  } else {
    Object.keys(state).forEach(function (key) {
      stack.push(key);
      makeStoreOperations(storeOperations, state[key], stack, key);
    });
  }
  stack.pop();
};

/**
 * An internal function that initializes the store, generates the API, and overloads actions
 * with the result of the operation. The latter is best understood if you consider two reducers
 * that both trigger something based on the same action type. In redux-operations, you can say
 * 1 action has many operations. Operation #2 may need the result of operation #1. By providing that information
 * in action.meta.operations.results, which is accessable in the resolve function, the developer can write
 * clean logic basic on previous results without the need for generators or middleware.
 */
var liftReducerWith = function liftReducerWith(reducer, initialCommittedState) {
  var initialLiftedState = {
    userState: initialCommittedState
  };

  return function () {
    var liftedState = arguments.length <= 0 || arguments[0] === undefined ? initialLiftedState : arguments[0];
    var liftedAction = arguments[1];
    var api = liftedState.api;
    var userState = liftedState.userState;

    if (liftedAction.type === INIT_REDUX || liftedAction.type === INIT_DEVTOOLS) {
      var initialResult = reducer(undefined, { type: INIT_REDUX_OPERATIONS });
      api = makeStoreAPI(initialResult);
    }
    var activeState = reducer(userState, liftedAction);
    var actionObject = api[liftedAction.type] || {};
    var operationArray = actionObject.operationArray;
    if (operationArray) {
      operationArray.forEach(function (operation) {
        var locationInState = operation.defaultLocation;
        // 3 possiblies: If a locationInState isn't given, use the default (for simple non-multi scenarios)
        // If Loc but no Name, or name == operation name, use given location (for dynamic or multi scenarios)
        // Otherwise, use default
        if (liftedAction.meta.operations.locationInState && (!liftedAction.meta.operations.operationName || operation.name === liftedAction.meta.operations.operationName)) {
          locationInState = liftedAction.meta.operations.locationInState;
        }
        var subState = walkState(locationInState, activeState);
        var newSubState = operation.resolve(subState, liftedAction);
        if (subState !== newSubState) {
          activeState = appendChangeToState(locationInState, activeState, newSubState);
        }
        liftedAction.meta.operations.results = liftedAction.meta.operations.results || {};
        liftedAction.meta.operations.results[operation.name] = { oldState: subState, state: newSubState };
      });
    }
    return {
      api: api,
      userState: activeState
    };
  };
};

/**
 * An internal function that overwrites the state with state.userState, so you can see the API in devtools
 * but you don't have to do anything differently in able to access your state.
 */
var unliftState = function unliftState(liftedState) {
  return liftedState.userState;
};

/**
 * An internal function that hands the dispatch function to every action allowing for pain-free
 * async actions (thunks, promises) that happen right in the resolve function.
 */
var unliftStore = function unliftStore(reduxOperationStore, liftReducer) {
  return _extends({}, reduxOperationStore, {
    dispatch: function dispatch(action) {
      action.meta = action.meta || {};
      action.meta.operations = action.meta.operations || {};
      action.meta.operations.dispatch = reduxOperationStore.dispatch;
      reduxOperationStore.dispatch(action);
      return action;
    },
    getState: function getState() {
      return unliftState(reduxOperationStore.getState());
    },
    replaceReducer: function replaceReducer(nextReducer) {
      reduxOperationStore.replaceReducer(liftReducer(nextReducer));
    }
  });
};

/**
 * An store enhancer factory that takes no arguments and returns a standard redux storeEnhancer that
 * works similarly to a very basic redux-devtools. It shows you the API of all your redux-operations reducers
 *
 * @returns {Function} redux storeEnhancer
 */
var reduxOperations = exports.reduxOperations = function reduxOperations() {
  return function (createStore) {
    return function (reducer, initialState, enhancer) {
      function liftReducer(r) {
        if (typeof r !== 'function') {
          throw new Error('Expected the reducer to be a function.');
        }
        return liftReducerWith(r, initialState);
      }

      var reduxOperationStore = createStore(liftReducer(reducer), enhancer);
      if (reduxOperationStore.reduxOperationStore) {
        throw new Error('redux-operations should not be applied more than once. Check your store configuration.');
      }
      reduxOperationStore.dispatch.toString = function () {
        return '<Dispatch Function>';
      };
      return unliftStore(reduxOperationStore, liftReducer);
    };
  };
};